#!/bin/bash

# This script aims to fix DWARF generated by AIX linker.
# For now, only TLS relocations are wrongly generatedy in 32bit.

VERSION=1.0
# Changelog:
# 2021/08/24 v1.0: Initial version

DEBUG=false
VERBOSE=true
TARGET_LIB=
TARGET_SYM=

# Make sure OpenSource will be taken first.
export PATH=/opt/freeware/bin:$PATH

function usage(){
cat  <<EOF
Script fixing wrong TLS relocations in a library.
Usage: fix-ld-dwarf.sh [OPTIONS] <target_lib>
Options:
  -d, --debug: debug mode
  -q, --quiet: quiet mode
  -s, --sym <sym>: Fix only relocations targeting <sym>
  -h, --help: Show this help
      --version: Show script version
EOF
}


# Get options
while [[ $# != 0 ]]; do
    case "$1" in
	--help | -h) usage; exit 0;;
	--version) echo $VERSION; exit 0;;
	--debug | -d) DEBUG=true; shift;;
	--quiet | -q) VERBOSE=FALSE; shift;;
	--sym | -s) TARGET_SYM=$2; shift; shift;;
	-*) echo "Unknown argument $1" >&2; usage; exit 1;;
	*) if [[ -n $TARGET_LIB ]]; then
	       echo "Please provide only one library." >&2
	       echo "Found $TARGET_LIB and $1" >&2
	       exit 1;
	   fi
	   TARGET_LIB=$1; shift;;
    esac
done

if [ -z $TARGET_LIB ]; then
    usage
    exit 1
fi

if file $TARGET_LIB | grep -q "64-bit XCOFF"; then
    echo "This script is only made for 32-bit shared objects." >&2
    exit 1
fi

dump_hv_f=/tmp/fixup$$-hv
dump_rv_f=/tmp/fixup$$-rv

dump -X32_64 -hv $TARGET_LIB > $dump_hv_f
dump -X32_64 -rv $TARGET_LIB > $dump_rv_f

function clean() {
    rm $dump_hv_f
    rm $dump_rv_f
}

# Return the number of relocation for a given section.
function get_section_nrel() {
    local sect=$1
    # Retrieve current section header.
    shdr=`cat "$dump_hv_f" | grep -A5 $sect`
    if [ -z "$shdr" ]; then
	# The section doesn't exist.
	return
    fi

    s_nrel=`echo "$shdr" | tail -1 | awk '{print $2}'`

    # A relocation number of 0xFFFF means that we must search for
    # the exact number inside one of the overflow sections.
    if [[ "$s_nrel" == "0xffff" ]]; then
	# First retrieve the number of the current section, using
	# the "Section Header for" anchor made by dump -hv.
	s_nb=`cat "$dump_hv_f" | grep "Section Header for" | grep -n $sect | cut -d: -f1`
	if [ -z "$s_nb" ]; then
	    echo "Cannot find section number for $sect" >&2
	    clean
	    exit 1
	fi

	# Retrieve the overflow section corresponding to the current
	# section by matching the section number with the overflow
	# s_nreloc/s_nlnno field.
	printf -v s_nb "0x%04x" $s_nb
	sofhdr=`cat "$dump_hv_f" | grep -A5 ".ovrflo" | grep -B5 "$s_nb"`
	if [ -z "$sofhdr" ]; then
	    echo "overflow section for $sect not found" >&2
	    clean
	    exit 1
	fi

	# The real relocation number is stored in s_paddr.
	s_nrel=`echo "$sofhdr" | head -3 | tail -1 | awk '{print $1}'`
    fi

    # Transform s_nrel into a decimal number for grep.
    printf "%d" $s_nrel
}

# Return the offset of the section's relocation.
function get_section_relptr() {
    local sect=$1
    # Retrieve current section header.
    shdr=`cat "$dump_hv_f" | grep -A5 $sect`
    if [ -z "$shdr" ]; then
	# The section doesn't exist.
	return
    fi

    echo "$shdr" | head -3 | tail -1 | awk '{print $5}'
}


# fix_rel_in_sec will search for the wrong relocation
# targetting the given symbol and will transform their
# r_type for R_TLS.
function fix_rel_in_sec() {
    local sect=$1
    local sym=$2

    s_nrel=`get_section_nrel $1`
    s_relptr=`get_section_relptr $1`

    # Find the wrong relocations for sym within this section.
    # dump -rv starts each section's part by "name:". Retrieve s_nrel lines
    # after that anchor will retrieve the relocations only made in this
    # section.
    wrong_rels=`cat "$dump_rv_f" | grep -A$s_nrel "$sect:" | grep -n "$sym\$"`

    if [[ -z $wrong_rels ]]; then
	return;
    fi

    [[ "$VERBOSE" == true ]] && \
	echo "Found" `echo "$wrong_rels" | wc -l` "relocation(s) to fix in $sect"
    [[ "$DEBUG" == "true" ]] && \
	echo "$wrong_rels"

    # For each relocation, use the line number retrieved by grep to
    # compute the address of the wrong relocation and transform it to
    # R_TLS.
    while read r; do
	r_nb=$((`echo "$r" | cut -d: -f1` - 2))
	ptr=$(($s_relptr + $r_nb*10 + 9))
	echo -n -e "\x20" | dd of=${TARGET_LIB} bs=1 count=1 seek=${ptr} conv=notrunc status=none
    done <<< "$wrong_rels"
}

function fix_rel_for_sym() {
    local sym=$1
    for s in .dwinfo .dwarnge .dwloc; do
	fix_rel_in_sec $s $sym
    done
}

function find_wrong_sym() {
    is_first=$1
    # First retrieve gcc errors.
    errors=`gcc -Wl,-bstatic $TARGET_LIB 2>&1`
    if [ -z "$errors" ]; then
	# Only exit when we are searching for the first symbol.
	# Otherwise, it could simply mean, everything has been fixed.
	if [[ $is_first == true ]]; then
	   echo "Cannot detect TLS symbols with wrong relocations:" >&2
	   echo "No error was triggered by \"gcc -Wl,-bstatic $TARGET_LIB\"." >&2
	   exit 0
	fi
	return
    fi

    if ! echo $errors | grep -q "ld: 0711-550 SEVERE ERROR"; then
	if [[ $is_first == true ]]; then
	    echo "Cannot detect TLS symbols with wrong relocations:" >&2
	    echo "Unknown error triggered by \"gcc -Wl,-bstatic $TARGET_LIB\"." >&2
	    exit 1
	fi
	return
    fi

    # Get relocation address and section with the line:
    # "RLD at address ... for section ... (...) references a thread-local".
    info_line=`echo "$errors" | head -2 | tail -1`
    rel_addr=`echo "$info_line" | awk '{print $4}'`
    printf -v rel_addr "0x%08x" $rel_addr
    rel_sect=`echo "$info_line" | awk '{print $8}' | sed "s/(//g" | sed "s/)//g"`

    # Retrieve symbol name.
    rel_sect_nrel=`get_section_nrel $rel_sect`
    cat "$dump_rv_f" | grep -A$rel_sect_nrel "$rel_sect:" | grep $rel_addr | awk '{print $7}'
}


if [[ -n $TARGET_SYM ]]; then
    [[ "$VERBOSE" == true ]] && \
	echo "Fixing relocations for symbol: $TARGET_SYM"
    fix_rel_for_sym $TARGET_SYM
else
    sym=`find_wrong_sym true`
    if [[ $? != 0 ]]; then exit 1; fi
    while [[ -n $sym ]]; do
	[[ "$VERBOSE" == true ]] && \
	    echo "Fixing relocations for symbol: $sym"
	fix_rel_for_sym $sym
	sym=`find_wrong_sym false`
	[[ "$VERBOSE" == true ]] && \
	    echo ""
    done
fi

clean

